@@ -50,9 +50,8 @@ using namespace llvm;
 ///
 /// @return Value*    The newly created induction variable for this loop.
 Value *createLoop(Value *LowerBound, Value *UpperBound, Value *Stride,
-                  PollyIRBuilder &Builder, Pass *P, LoopInfo &LI,
-                  DominatorTree &DT, BasicBlock *&ExitBlock,
-                  ICmpInst::Predicate Predicate,
+                  PollyIRBuilder &Builder, LoopInfo &LI, DominatorTree &DT,
+                  BasicBlock *&ExitBlock, ICmpInst::Predicate Predicate,
                   ScopAnnotator *Annotator = NULL, bool Parallel = false,
                   bool UseGuard = true);
 
@@ -99,9 +98,9 @@ Value *createLoop(Value *LowerBound, Value *UpperBound, Value *Stride,
 class ParallelLoopGenerator {
 public:
   /// Create a parallel loop generator for the current function.
-  ParallelLoopGenerator(PollyIRBuilder &Builder, Pass *P, LoopInfo &LI,
+  ParallelLoopGenerator(PollyIRBuilder &Builder, LoopInfo &LI,
                         DominatorTree &DT, const DataLayout &DL)
-      : Builder(Builder), P(P), LI(LI), DT(DT), DL(DL),
+      : Builder(Builder), LI(LI), DT(DT),
         LongType(
             Type::getIntNTy(Builder.getContext(), DL.getPointerSizeInBits())),
         M(Builder.GetInsertBlock()->getParent()->getParent()) {}
@@ -131,18 +130,12 @@ private:
   /// The IR builder we use to create instructions.
   PollyIRBuilder &Builder;
 
-  /// A pass pointer to update analysis information.
-  Pass *P;
-
   /// The loop info of the current function we need to update.
   LoopInfo &LI;
 
   /// The dominance tree of the current function we need to update.
   DominatorTree &DT;
 
-  /// The target layout to get the right size for types.
-  const DataLayout &DL;
-
   /// The type of a "long" on this hardware used for backend calls.
   Type *LongType;
 
@@ -179,7 +172,7 @@ public:
   /// Create a runtime library call to allow cleanup of the thread.
   ///
   /// @note This function is called right before the thread will exit the
-  ///       subfunction and only if the runtime system depends depends on it.
+  ///       subfunction and only if the runtime system depends on it.
   void createCallCleanupThread();
 
   /// Create a struct for all @p Values and store them in there.
